#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Drawing;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Problems.GeneralizedQuadraticAssignment {
  [Item("Generalized Quadratic Assignment Problem", "The Generalized Quadratic Assignment Problem (GQAP) is a generalization of the QAP in that it allows to assign multiple facilities (here called equipment) to a single location. The problem is described in Lee, C.G., and Ma, Z. 2003. The Generalized Quadratic Assignment Problem. Technical Report M5S 3G8, University of Toronto, Canada.")]
  [Creatable("Problems")]
  [StorableClass]
  public sealed class GeneralizedQuadraticAssignmentProblem : SingleObjectiveHeuristicOptimizationProblem<IGQAPEvaluator, IGQAPSolutionCreator>, IStorableContent {
    public override Image ItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Type; }
    }

    public string Filename { get; set; }

    #region Parameter Properties
    public ValueParameter<DoubleMatrix> WeightsParameter {
      get { return (ValueParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ValueParameter<DoubleMatrix> DistancesParameter {
      get { return (ValueParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ValueParameter<DoubleMatrix> InstallationCostsParameter {
      get { return (ValueParameter<DoubleMatrix>)Parameters["InstallationCosts"]; }
    }
    public ValueParameter<DoubleArray> DemandsParameter {
      get { return (ValueParameter<DoubleArray>)Parameters["Demands"]; }
    }
    public ValueParameter<DoubleArray> CapacitiesParameter {
      get { return (ValueParameter<DoubleArray>)Parameters["Capacities"]; }
    }
    public FixedValueParameter<DoubleValue> TransportationCostsParameter {
      get { return (FixedValueParameter<DoubleValue>)Parameters["TransportationCosts"]; }
    }
    public FixedValueParameter<DoubleValue> OverbookedCapacityPenaltyParameter {
      get { return (FixedValueParameter<DoubleValue>)Parameters["OverbookedCapacityPenalty"]; }
    }
    public OptionalValueParameter<IItem> BestKnownSolutionParameter {
      get { return (OptionalValueParameter<IItem>)Parameters["BestKnownSolution"]; }
    }
    #endregion

    #region Properties
    public DoubleMatrix Weights {
      get { return WeightsParameter.Value; }
      set { WeightsParameter.Value = value; }
    }
    public DoubleMatrix Distances {
      get { return DistancesParameter.Value; }
      set { DistancesParameter.Value = value; }
    }
    public DoubleMatrix InstallationCosts {
      get { return InstallationCostsParameter.Value; }
      set { InstallationCostsParameter.Value = value; }
    }
    public DoubleArray Demands {
      get { return DemandsParameter.Value; }
      set { DemandsParameter.Value = value; }
    }
    public DoubleArray Capacities {
      get { return CapacitiesParameter.Value; }
      set { CapacitiesParameter.Value = value; }
    }
    public double TransportationCosts {
      get { return TransportationCostsParameter.Value.Value; }
      set { TransportationCostsParameter.Value.Value = value; }
    }
    public double OverbookedCapacityPenalty {
      get { return TransportationCostsParameter.Value.Value; }
      set { TransportationCostsParameter.Value.Value = value; }
    }
    #endregion

    [Storable]
    private BestGQAPSolutionAnalyzer bestSolutionAnalyzer;
    public BestGQAPSolutionAnalyzer BestSolutionAnalyzer {
      get { return bestSolutionAnalyzer; }
      set { bestSolutionAnalyzer = value; }
    }

    [StorableConstructor]
    private GeneralizedQuadraticAssignmentProblem(bool deserializing) : base(deserializing) { }
    private GeneralizedQuadraticAssignmentProblem(GeneralizedQuadraticAssignmentProblem original, Cloner cloner)
      : base(original, cloner) {
      bestSolutionAnalyzer = cloner.Clone(original.bestSolutionAnalyzer);
      AttachEventHandlers();
    }
    public GeneralizedQuadraticAssignmentProblem()
      : base(new GQAPEvaluator(), new RandomSolutionCreator()) {
      Parameters.Add(new ValueParameter<DoubleMatrix>("Weights", "The weights matrix describes the flows between the equipments.", new DoubleMatrix()));
      Parameters.Add(new ValueParameter<DoubleMatrix>("Distances", "The distances matrix describes the distances between the locations at which the equipment can be installed.", new DoubleMatrix()));
      Parameters.Add(new ValueParameter<DoubleMatrix>("InstallationCosts", "The installation costs matrix describes the installation costs of installing equipment i at location j", new DoubleMatrix()));
      Parameters.Add(new FixedValueParameter<DoubleValue>("TransportationCosts", "The transportation cost represents the flow-unit per distance-unit cost factor. This value can also be set to 1 if these costs are factored into the weights matrix already.", new DoubleValue(1)));
      Parameters.Add(new FixedValueParameter<DoubleValue>("OverbookedCapacityPenalty", "The multiplier for the constraint violation when added to the quality.", new DoubleValue(1000)));
      Parameters.Add(new ValueParameter<DoubleArray>("Demands", "The demands vector describes the space requirements of the equipments.", new DoubleArray()));
      Parameters.Add(new ValueParameter<DoubleArray>("Capacities", "The capacities vector describes the available space at the locations.", new DoubleArray()));
      Parameters.Add(new OptionalValueParameter<IItem>("BestKnownSolution", "The best known solution (if available)", null));
      Parameters.Add(new OptionalValueParameter<StringArray>("EquipmentNames", "Optional: A list of names that describes the equipments.", null, false));
      Parameters.Add(new OptionalValueParameter<StringArray>("LocationNames", "Optional: A list of names that describes the locations.", null, false));

      WeightsParameter.ReactOnValueToStringChangedAndValueItemImageChanged = false;
      DistancesParameter.ReactOnValueToStringChangedAndValueItemImageChanged = false;
      InstallationCostsParameter.ReactOnValueToStringChangedAndValueItemImageChanged = false;

      Weights = new DoubleMatrix(5, 5);
      Weights[0, 0] = Weights[1, 1] = Weights[2, 2] = Weights[3, 3] = Weights[4, 4] = 0;
      Weights[0, 1] = Weights[1, 0] = Weights[0, 2] = Weights[2, 0] = Weights[0, 3] = Weights[3, 0] = Weights[0, 4] = Weights[4, 0] = 10;
      Weights[1, 2] = Weights[2, 1] = Weights[1, 3] = Weights[3, 1] = Weights[1, 4] = Weights[4, 1] = 5;
      Weights[2, 3] = Weights[3, 2] = Weights[2, 4] = Weights[4, 2] = 7.5;
      Weights[3, 4] = Weights[4, 3] = 2.5;

      Distances = new DoubleMatrix(3, 3);
      Distances[0, 0] = Distances[1, 1] = Distances[2, 2] = 0;
      Distances[0, 1] = Distances[1, 0] = Distances[1, 2] = Distances[2, 1] = 1;
      Distances[0, 2] = Distances[2, 0] = 2;

      InstallationCosts = new DoubleMatrix(5, 3);

      TransportationCosts = 1;

      Demands = new DoubleArray(5);
      Demands[0] = 2; Demands[1] = 1; Demands[2] = 3; Demands[3] = 1; Demands[4] = 1;

      Capacities = new DoubleArray(3);
      Capacities[0] = 4; Capacities[1] = 1; Capacities[2] = 4;

      SolutionCreator.AssignmentParameter.ActualName = "Assignment";
      Parameterize();

      InitializeOperators();
      AttachEventHandlers();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GeneralizedQuadraticAssignmentProblem(this, cloner);
    }

    #region Events
    protected override void OnOperatorsChanged() {
      base.OnOperatorsChanged();
      Parameterize();
    }
    protected override void OnEvaluatorChanged() {
      base.OnEvaluatorChanged();
      Parameterize();
      Evaluator.QualityParameter.ActualNameChanged += new System.EventHandler(Evaluator_QualityParameter_ActualNameChanged);
    }
    protected override void OnSolutionCreatorChanged() {
      base.OnSolutionCreatorChanged();
      Parameterize();
      SolutionCreator.AssignmentParameter.ActualNameChanged += new EventHandler(SolutionCreator_IntegerVectorParameter_ActualNameChanged);
    }

    private void Evaluator_QualityParameter_ActualNameChanged(object sender, EventArgs e) {
      Parameterize();
    }
    private void SolutionCreator_IntegerVectorParameter_ActualNameChanged(object sender, EventArgs e) {
      Parameterize();
    }
    #endregion

    #region Helpers
    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      AttachEventHandlers();
    }

    private void AttachEventHandlers() {
      Evaluator.QualityParameter.ActualNameChanged += new System.EventHandler(Evaluator_QualityParameter_ActualNameChanged);
      SolutionCreator.AssignmentParameter.ActualNameChanged += new EventHandler(SolutionCreator_IntegerVectorParameter_ActualNameChanged);
    }

    private void InitializeOperators() {
      Operators.AddRange(ApplicationManager.Manager.GetInstances<IIntegerVectorOperator>());
      Operators.AddRange(ApplicationManager.Manager.GetInstances<IGQAPOperator>());
      Operators.RemoveAll(x => x is ISingleObjectiveMoveEvaluator);
      Operators.AddRange(ApplicationManager.Manager.GetInstances<IGQAPMoveEvaluator>());
      Operators.Add(new BestGQAPSolutionAnalyzer());
      Parameterize();
    }

    private void Parameterize() {
      Evaluator.WeightsParameter.ActualName = WeightsParameter.Name;
      Evaluator.DistancesParameter.ActualName = DistancesParameter.Name;
      Evaluator.InstallationCostsParameter.ActualName = InstallationCostsParameter.Name;
      Evaluator.TransportationCostsParameter.ActualName = TransportationCostsParameter.Name;
      Evaluator.DemandsParameter.ActualName = DemandsParameter.Name;
      Evaluator.CapacitiesParameter.ActualName = CapacitiesParameter.Name;
      Evaluator.AssignmentParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;

      SolutionCreator.DemandsParameter.ActualName = DemandsParameter.Name;
      SolutionCreator.CapacitiesParameter.ActualName = CapacitiesParameter.Name;

      foreach (var op in Operators.OfType<IEquipmentAwareGQAPOperator>()) {
        op.DemandsParameter.ActualName = DemandsParameter.Name;
      }
      foreach (var op in Operators.OfType<IGQAPCrossover>()) {
        op.ParentsParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
        op.ChildParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
      }
      foreach (var op in Operators.OfType<IGQAPEvaluationOperator>()) {
        op.AssignmentParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
        op.TransportationCostsParameter.ActualName = TransportationCostsParameter.Name;
        op.OverbookedCapacityPenaltyParameter.ActualName = OverbookedCapacityPenaltyParameter.Name;
        op.WeightsParameter.ActualName = WeightsParameter.Name;
        op.DistancesParameter.ActualName = DistancesParameter.Name;
        op.InstallationCostsParameter.ActualName = InstallationCostsParameter.Name;
        op.DemandsParameter.ActualName = DemandsParameter.Name;
        op.CapacitiesParameter.ActualName = CapacitiesParameter.Name;
      }
      foreach (var op in Operators.OfType<IGQAPLocalImprovementOperator>()) {
        op.AssignmentParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
        op.QualityParameter.ActualName = Evaluator.QualityParameter.ActualName;
        op.FlowDistanceQualityParameter.ActualName = Evaluator.FlowDistanceQualityParameter.ActualName;
        op.InstallationQualityParameter.ActualName = Evaluator.InstallationQualityParameter.ActualName;
        op.OverbookedCapacityParameter.ActualName = Evaluator.OverbookedCapacityParameter.ActualName;
        op.TransportationCostsParameter.ActualName = TransportationCostsParameter.Name;
        op.OverbookedCapacityPenaltyParameter.ActualName = OverbookedCapacityPenaltyParameter.Name;
        op.WeightsParameter.ActualName = WeightsParameter.Name;
        op.DistancesParameter.ActualName = DistancesParameter.Name;
        op.InstallationCostsParameter.ActualName = InstallationCostsParameter.Name;
        op.DemandsParameter.ActualName = DemandsParameter.Name;
        op.CapacitiesParameter.ActualName = CapacitiesParameter.Name;
      }
      foreach (var op in Operators.OfType<IGQAPManipulator>()) {
        op.IntegerVectorParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
      }
      foreach (var op in Operators.OfType<IGQAPMerger>()) {
        op.AssignmentParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
        op.QualityParameter.ActualName = Evaluator.QualityParameter.ActualName;
        op.MaximizationParameter.ActualName = MaximizationParameter.Name;
      }
      foreach (var op in Operators.OfType<IGQAPMoveOperator>()) {
        op.AssignmentParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
      }
      foreach (var op in Operators.OfType<ILocationAwareGQAPOperator>()) {
        op.CapacitiesParameter.ActualName = CapacitiesParameter.Name;
      }

      foreach (var op in Operators.OfType<IIntegerVectorCrossover>()) {
        op.ParentsParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
        op.ChildParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
      }
      foreach (var op in Operators.OfType<IIntegerVectorManipulator>()) {
        op.IntegerVectorParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
      }

      if (BestSolutionAnalyzer != null) {
        BestSolutionAnalyzer.QualityParameter.ActualName = Evaluator.QualityParameter.ActualName;
        BestSolutionAnalyzer.DistancesParameter.ActualName = DistancesParameter.Name;
        BestSolutionAnalyzer.WeightsParameter.ActualName = WeightsParameter.Name;
        BestSolutionAnalyzer.AssignmentParameter.ActualName = SolutionCreator.AssignmentParameter.ActualName;
        BestSolutionAnalyzer.ResultsParameter.ActualName = "Results";
        BestSolutionAnalyzer.BestKnownQualityParameter.ActualName = BestKnownQualityParameter.Name;
        BestSolutionAnalyzer.MaximizationParameter.ActualName = MaximizationParameter.Name;
      }
    }
    #endregion
  }
}
